#!/usr/bin/env python3

__version__ = '1.13.2'

#
# WELCOME!
# --------
#
# If you are reading this file you are probably a Software Engineer looking for
# some examples of how to use the VirusTotal API for private file scanning. You
# are in the right place, this file is for you.
#
# This little program will guide you through the steps required for scanning a
# file privately in VirusTotal, and will teach you how to extract some useful
# information from the generated reports. Ready to start?
#
# REQUIREMENTS
# ------------
#
# This program has very little requirements, but still it has some. The first
# one is Python 3, of course, because this is a Python 3 program. Python 2 is
# not supported.
#
# It also depends on some libraries, some of them are standard libraries that
# come already with Python, like the following ones:
#
# EXAMPLE USAGE
#
#  1. Scan a file: ./privscan.py --network-enabled installer.exe
#  2. Open URL in explorer:
#     echo "explorer http://youtube.com" > youtube.bat && \
#         ./privscan.py  --network-enabled  youtube.bat

import argparse
import base64
import datetime
import hashlib
import json
import os
import re
import time
import sys
import urllib3

#
# But the program also depends on third-party libraries that must be installed
# separately. For installing those libraries you can use:
#
# pip3 install requests rich
#
# But don't worry, if you try to run this program without installing the required
# libraries it will tell you what to install.
#
try:
  import requests
  from rich.console import Console
  from rich.prompt import Confirm
  from rich.prompt import Prompt
  from rich.table import Table
  from rich.text import Text
  from rich.tree import Tree
  from rich.style import Style
  from rich import box
except ModuleNotFoundError:
  print('\nATENTION!\n')
  print('Some of the libraries required by this program are not currently installed.')
  print('Please run "pip3 install rich requests" for installing them.')
  sys.exit(1)

# Disable warning in case the user specifies the --skip-ssl-verification flag.
requests.packages.urllib3.disable_warnings(
    requests.packages.urllib3.exceptions.InsecureRequestWarning)

#
# GLOBALS
# -------
#
# Here comes some global definitions. Nothing important here, just a few things
# that are used in multiple places of the code.
#
SPINNER = {'spinner': 'dots', 'spinner_style': 'bright_yellow'}
LEFT_COLUMN_MIN_WIDTH = 15
CONSOLE_WITH = 130

console = Console(width=CONSOLE_WITH)

#
# VERSION CHECK
# -------------
#
# If there is a new version of the script it will automatically update it.
#
def version_update(verify_ssl_cert=True):
  response = requests.get(
      'https://storage.googleapis.com/vtcdn/api-scripts/privscan.py',
      verify=verify_ssl_cert)
  if response.status_code != 200:
    console.print('Unable to check the last version of the script')
    return False

  content = response.content.decode()
  last_version = re.search(r'^__version__ = \'([^\']*)\'$', content, re.M)
  if last_version and last_version.group(1) == __version__:
    return False

  console.print('New version available, updating script...')
  try:
    with open(__file__, 'w', encoding='utf-8') as f:
      f.write(content)
  except Exception as e:
    console.print('Error updating the script: %s' % e)
    return False
  return True

def proxy_check():
  try:
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
    requests.get('https://www.virustotal.com', verify=True)
    return False
  except requests.exceptions.SSLError:
    return True

#
# UTILITY FUNCTIONS AND CLASSES
# -----------------------------
#
# Here we start defining some utility functions and classes that are going to
# be useful later on. One of such functions is `get_key`, which allows to extract
# a specific field from a dictionary that contains other nested dictionaries.
#
# For example, given the following dictionary:
#
#  {
#    'foo': {
#      'bar': {
#        'baz': <some value>
#      }
#    }
#  }
#
#  get_key will return <some value> when asked for the key "foo.bar.baz" where
#  the dot (.) is used as key separator. If key is "foo.bar" the result is:
#
#  {
#    'baz': <some value>
#  }
#
#  This function is very useful because the JSON responses returned by the
#  VirusTotal API usually consist on heavily-nested dictionaries, and get_key
#  simplifies the process of extracting values from them.
#
def get_key(dictionary, key, default_value=None):
  """Get value from nested dictionaries.

  Args:
    dictionary: A dictionary with possibly nested dictionaries.
    key: A string with the full key to retrieve.
    default_value: Value returned if the key is not found.

  Returns:
    The value stored in the provided key.

  Raises:
    KeyError: if the provided key is not valid.
  """
  dictionary = dictionary or {}
  keys = key.split('.')
  field_name = keys.pop()
  for k in keys:
    if k not in dictionary:
      return default_value
    dictionary = dictionary[k]
    if not isinstance(dictionary, dict):
      raise KeyError('"%s" is not a dictionary' % k)
  return dictionary.get(field_name, default_value)
#
# Generates object ID for an URL. The ID generated can be used in API calls that
# expect an URL ID like GET /urls/<id>.
#
def url_id(url):
  return base64.urlsafe_b64encode(url.encode()).decode().strip("=")
#
# ApiClient is another helper class used in this program. It's a very simple
# class that facilitates making GET and POST requests to the VirusTotal API.
#
# This class sets the necessary HTTP headers on each request, like for
# example the "X-Apikey" header that contains your API key and allows VirusTotal
# identifying the user making the call.
#
# It also sets the "Accept-Encoding: gzip" header, letting know the VirusTotal
# API servers that they can respond with gzip-compressed data, reducing the
# used bandwith and the latency of your requests. Finally, it also sets the
# "User-Agent" agent header to "Private Scan Sample Script; gzip". The important
# detail about the user agent is the "; gzip" part at the end. This is an
# additional requirement by Google AppEngine (the service that hosts the
# VirusTotal API), which refuses to send gzip-compressed to clients that don't
# send the "gzip" string somewhere in the user agent. For more details see:
# https://stackoverflow.com/questions/8471681/appengine-gzip-compressing
#
class ApiClient:
  def __init__(self, apikey, verify=True):
    self._apikey = apikey
    self._host = 'https://www.virustotal.com'
    self._verify = verify
    self._headers = {
        'X-Apikey': self._apikey,
        'Accept-Encoding': 'gzip',
        'User-Agent': 'Private Scan Sample Script; gzip'}

  def get(self, path, **kwargs):
    if path.startswith('http'):
      url = path
    else:
      url = self._host + path
    if 'timeout' not in kwargs:
      kwargs['timeout'] = 20  # Set default timeout to 20 seconds.
    response = requests.get(
        url, headers=self._headers, verify=self._verify, **kwargs)
    if response.status_code != 200:
      raise Exception('Error %d: %s' % (response.status_code, response.text))
    return response

  def post(self, path, **kwargs):
    if path.startswith('http'):
      url = path
    else:
      url = self._host + path
    if 'timeout' not in kwargs:
      kwargs['timeout'] = 20  # Set default timeout to 20 seconds.
    response = requests.post(
        url, headers=self._headers, verify=self._verify, **kwargs)
    if response.status_code != 200:
      raise Exception('Error %d: %s' % (response.status_code, response.text))
    return response
#
# compute_sha256 is a small function that simply calculates the SHA-256 of
# a local file. We use the file's SHA-256 for checking if you already scanned
# the file privately in the past.
#
def compute_sha256(file_path):
  with open(file_path, 'rb') as fin:
    file_hash = hashlib.sha256(fin.read()).hexdigest()
  return file_hash
#
# PROGRAM LOGIC
# -------------
#
# Now follows a series of functions that print information about files. All of
# these function receive a "file_info" argument consisting in the response
# obtained from the API when we ask about a file. This dictionary has the
# following structure:
#
#  {
#    'data': {
#      'id': <file's SHA-256>,
#      'type': 'file',
#      'attributes': {
#           ... file's attributes
#      }
#    }
#  }
#
# That's the overall structure of a file object, you can find more information
# about VirustTotal API objects in:
#
# https://docs.virustotal.com/reference/objects
#
# Additionally, you can obtain a more detailed description of all the attributes
# that can be found in a file object by visiting:
#
# https://docs.virustotal.com/reference/files
#
# This first function "print_file_identification" prints general information
# about a file, like its hashes, type, size, etc.
#
def print_file_identification(file_info):
  attributes = get_key(file_info, 'data.attributes', {})

  console.rule('[yellow2]File identification', align='left')

  table = Table(show_header=False, box=box.SIMPLE)
  table.add_column(justify='right', min_width=LEFT_COLUMN_MIN_WIDTH)
  table.add_column(style='yellow')

  table.add_row('MD5', attributes.get('md5'))
  table.add_row('SHA-1', attributes.get('sha1'))
  table.add_row('SHA-256', attributes.get('sha256'))
  table.add_row('File size', str(attributes.get('size')))
  table.add_row('File type', attributes.get('type_description'))
  table.add_row('Magic', attributes.get('magic'))
  table.add_row('Tags', ' '.join(attributes.get('tags',[])))

  console.print(table)

#
# print_digital_signature prints information about the file's digital signature
# (if any). This information is found in an attribute named "signature_info",
# which is a dictionary with additional fields.
#
def print_digital_signature(file_info):
    signature_info = get_key(file_info, 'data.attributes.signature_info', {})

    if not signature_info:
      return

    table = Table(show_header=False, box=box.SIMPLE)
    table.add_column(justify='right', min_width=LEFT_COLUMN_MIN_WIDTH)
    table.add_column(style='yellow')

    description = signature_info.get('description', '').strip()
    signers = signature_info.get('signers', '').strip()

    if description:
      table.add_row('Description', description)

    if signers:
      table.add_row('Signers', signers, end_section=True)

    verification_msg = signature_info.get('verified')

    if verification_msg == 'Signed':
      table.add_row('', 'File signed with valid signature', style='green')
    else:
      table.add_row('', 'No valid signature found', style='red')

    if description or signers or verification_msg:
      console.rule('[yellow2]File signature', align='left')
      console.print(table)
#
# print_crowdsourced_yara prints information about whether the file has
# matched some YARA rules that contributed by community. This information
# is found in the "crowdsourced_yara_results", in the form of a list of
# dictionaries with fields like "rule_name", "description", "author", etc.
#
def print_crowdsourced_yara(file_info):
  crowdsourced_yara_results = get_key(
      file_info, 'data.attributes.crowdsourced_yara_results', [])

  if not crowdsourced_yara_results:
    return

  console.rule('[yellow2]Crowdsourced YARA rule matches', align='left')

  table = Table(show_header=False, box=box.SIMPLE)
  table.add_column(justify='right', min_width=LEFT_COLUMN_MIN_WIDTH)
  table.add_column(style='yellow')

  for r in crowdsourced_yara_results:
    description = r.get('description')
    if description:
      table.add_row('Description', description)
    table.add_row('Ruleset', r['ruleset_name'])
    table.add_row('Rule', r['rule_name'])
    table.add_row('Author', r['author'])
    table.add_row('Source', r['source'], end_section=True)

  console.print(table)
#
# print_crowdsourced_sigma is similar to print_crowdsourced_yara, but it
# prints information about Sigma rules. You can find more information about
# Sigma in https://github.com/SigmaHQ/sigma, but in a nutshell, Sigma allows
# to write detection rules based in log files.
#
def print_crowdsourced_sigma(file_info):
  matches = get_key(file_info, 'data.attributes.sigma_analysis_results', [])

  if not matches:
    return

  console.rule('[yellow2]Crowdsourced Sigma rule matches ', align='left')

  table = Table(show_header=False, box=box.SIMPLE)
  table.add_column(justify='right', min_width=LEFT_COLUMN_MIN_WIDTH)
  table.add_column(style='yellow')

  for r in matches:
    level = r['rule_level']

    if level == 'high':
      style = 'red'
    elif level == 'medium':
      style = 'orange'
    else:
      style = 'green'

    table.add_row('Severity', Text(r['rule_level'].capitalize(), style=style))
    table.add_row('Description', r['rule_description'])
    table.add_row('Rule', r['rule_title'])
    table.add_row('Author', r['rule_author'])
    table.add_row('Source', r['rule_source'], end_section=True)

  console.print(table)


def print_crowdsourced_ids(file_info):
  crowdsourced_ids_results = get_key(
      file_info, 'data.attributes.crowdsourced_ids_results', [])

  if not crowdsourced_ids_results:
    return

  console.rule('[yellow2]Crowdsourced IDS rule matches', align='left')

  table = Table(show_header=False, box=box.SIMPLE)
  table.add_column(justify='right', min_width=LEFT_COLUMN_MIN_WIDTH)
  table.add_column(style='yellow')

  for r in crowdsourced_ids_results:
    level = r['alert_severity']

    if level == 'high':
      style = 'red'
    elif level == 'medium':
      style = 'orange'
    else:
      style = 'green'

    table.add_row('Severity', Text(r['alert_severity'].capitalize(), style=style))
    table.add_row('Description', r['rule_msg'])
    table.add_row('Source', r['rule_source'], end_section=True)

  console.print(table)


def print_sandbox_detonation_verdicts(file_info):
  verdicts = get_key(
      file_info, 'data.attributes.sandbox_verdicts', {})

  if not verdicts:
    return

  console.rule('[yellow2]Sandbox detonation verdics', align='left')

  table = Table(show_header=False, box=box.SIMPLE)
  table.add_column(justify='right', min_width=LEFT_COLUMN_MIN_WIDTH)
  table.add_column()

  for sandbox, verdicts in verdicts.items():
    l = [
      '[red]%s[/red]' % v if v != 'CLEAN' else '[green]%s[/green]' % v
      for v in verdicts.get('malware_classification', [])]

    table.add_row(sandbox, '\n'.join(l))

  console.print(table)

  malware_config = get_key(
      file_info, 'data.attributes.malware_config', {})

  if not malware_config:
    return

  table = Table(show_header=False, box=box.SIMPLE)
  table.add_column(justify='right', min_width=LEFT_COLUMN_MIN_WIDTH)
  table.add_column(style='yellow')

  if 'c2urls' in malware_config:
    table.add_row((
        '%s command & control' % malware_config.get('family', '')
      ).strip().capitalize(),
      '\n'.join(malware_config['c2urls']))

  console.print(table)

def print_sandbox_detonation_summary(behaviour_info, netloc_info):

  console.rule('[yellow2]Sandbox detonation summary', align='left')

  table = Table(show_header=False, box=box.SIMPLE)
  table.add_column(justify='right', min_width=LEFT_COLUMN_MIN_WIDTH)
  table.add_column(style='yellow')

  table.add_row('Files dropped',
      str(len(get_key(behaviour_info, 'data.files_dropped', []))))

  table.add_row('Files opened',
      str(len(get_key(behaviour_info, 'data.files_opened', []))))

  table.add_row('Files written',
      str(len(get_key(behaviour_info, 'data.files_written', []))))

  table.add_row('Registry keys opened',
      str(len(get_key(behaviour_info, 'data.registry_keys_opened', []))))

  table.add_row('Registry keys set',
      str(len(get_key(behaviour_info, 'data.registry_keys_set', []))))

  table.add_row('Registry keys deleted',
      str(len(get_key(behaviour_info, 'data.registry_keys_deleted', []))))

  table.add_row('Mutexes opened',
      str(len(get_key(behaviour_info, 'data.mutexes_opened', []))))

  table.add_row('Mutexes created',
      str(len(get_key(behaviour_info, 'data.mutexes_created', []))))

  console.print(table)

  def _build_tree(tree, processes):
    for process in processes:
      _build_tree(
          tree.add(process['name'].strip('" ')),
          process.get('children', []))

  processes = get_key(behaviour_info, 'data.processes_tree', [])
  if processes:
    table = Table("Processes tree", box=box.SIMPLE, width=CONSOLE_WITH)
    table.add_column()

    for process in processes:
      tree = Tree(process['name'].strip('" '))
      _build_tree(tree, process.get('children', []))
      table.add_row(tree)

    console.print(table)

  dns_lookups = get_key(behaviour_info, 'data.dns_lookups', [])
  http_conversations = get_key(behaviour_info, 'data.http_conversations', [])
  ip_traffic = get_key(behaviour_info, 'data.ip_traffic', [])

  if not dns_lookups and not http_conversations and not ip_traffic:
    return

  table = Table(box=box.SIMPLE, width=CONSOLE_WITH)
  table.add_column('Network communications')
  table.add_column('Summary')
  table.add_column('Detection ratio', justify='right')

  for item in dns_lookups:
    hostname = item['hostname']
    malicious = netloc_info[hostname]['malicious']
    harmless = netloc_info[hostname]['harmless']
    undetected = netloc_info[hostname]['undetected']
    table.add_row(
        hostname,
        netloc_info[hostname]['summary'],
        Text(
            '%d / %d' % (malicious, malicious + harmless + undetected),
            style='red' if malicious else 'green'))

  for item in ip_traffic:
    ip_address = item['destination_ip']
    malicious = netloc_info[ip_address]['malicious']
    harmless = netloc_info[ip_address]['harmless']
    undetected = netloc_info[ip_address]['undetected']
    table.add_row(
        ip_address,
        netloc_info[ip_address]['summary'],
        Text(
            '%d / %d' % (malicious, malicious + harmless + undetected),
            style='red' if malicious else 'green'))

  for item in http_conversations:
    url = item['url']
    malicious = netloc_info[url]['malicious']
    harmless = netloc_info[url]['harmless']
    undetected = netloc_info[url]['undetected']
    table.add_row(
        url,
        netloc_info[url]['summary'],
        Text(
            '%d / %d' % (malicious, malicious + harmless + undetected),
            style='red' if malicious else 'green'))

  console.print(table)


def print_mitre_attack_tree(mitre_tree):

  console.rule('[yellow2]MITRE ATT&CK', align='left')

  tree = Tree('')
  for sandbox_name, data in mitre_tree['data'].items():
    if not data['tactics']:
      continue
    sandbox_node = tree.add(sandbox_name)
    for tactic in data['tactics']:
      t_id = tactic['id']
      t_name = tactic['name']
      t_link = tactic['link']
      tactic_node = sandbox_node.add(f'[link {t_link}]{t_id}: {t_name}')
      for technique in tactic['techniques']:
        t_id = technique['id']
        t_link = technique['link']
        t_name = technique['name']
        technique_node = tactic_node.add(f'[link {t_link}]{t_id}: {t_name}')
        label = ''
        for signature in technique['signatures']:
          if not signature.get('description'):
            continue
          label = ''
          if signature['severity'] == 'HIGH':
            label = '[red]❗️'
          technique_node.add(label + signature['description'])

  console.print(tree)
  console.print()

def print_sandbox_detonation_details(behavior_reports):

  table = Table(box=box.SIMPLE, show_footer=False, width=CONSOLE_WITH)
  table.add_column('Domain memory patterns')
  rows = 0

  for report in  behavior_reports.get('data', []):
    for domain in get_key(report, 'attributes.memory_pattern_domains', []):
      rows += 1
      table.add_row(domain)

  if rows:
    console.print(table)

  table = Table(box=box.SIMPLE, show_footer=False, width=CONSOLE_WITH)
  table.add_column('URL memory patterns')
  rows = 0

  for report in  behavior_reports.get('data', []):
    for url in get_key(report,'attributes.memory_pattern_urls', []):
      rows += 1
      table.add_row(url)

  if rows:
    console.print(table)

  table = Table(box=box.SIMPLE, show_footer=False, width=CONSOLE_WITH)
  table.add_column('IP memory patterns')
  rows = 0

  for report in  behavior_reports.get('data', []):
    for ip in get_key(report,'attributes.memory_pattern_ips', []):
      rows += 1
      table.add_row(ip)

  if rows:
    console.print(table)


def print_files_table(files_info, table_title, link=None):
  table = Table(box=box.SIMPLE, show_footer=False, width=CONSOLE_WITH)
  table.add_column(table_title)
  table.add_column('Type')
  table.add_column('Detection ratio', justify='right')

  for file in files_info:
    analysis_stats =  get_key(file, 'attributes.last_analysis_stats')
    malicious = analysis_stats['malicious']
    undetected = analysis_stats['undetected']
    table.add_row(
        file['id'],
        get_key(file, 'attributes.type_description'),
        Text(
            '%d / %d' % (malicious, malicious + undetected),
            style='red' if malicious else 'green'))

  console.print(table)

  if link:
    table = Table(box=box.SIMPLE, show_header=False, width=CONSOLE_WITH)
    table.add_row('See full list at:', style='dim')
    table.add_row('[link={link}]{link}[/link]'.format(link=link))
    console.print(table)

def print_unfinished_sandboxes(analysis):
  """Prints the reason why any sandbox didn't finish its execution."""
  sandbox_status = get_key(analysis, 'data.attributes.sandbox_status', {})
  for sandbox_name, data in sandbox_status.items():
    if data['status'] != 'finished':
      message = (
          '[red]✘[/red] Unfinished sandbox execution of {} with status '
          '"{}"'.format(sandbox_name, data['status']))
      if data.get('message'):
        message += ': {}'.format(data['message'])
      console.print(message)


def print_embedded_domains(embedded_domains):
  embedded_domains = get_key(embedded_domains, 'data', [])

  table = Table(box=box.SIMPLE, show_footer=False, width=CONSOLE_WITH)
  table.add_column('Embedded domains')
  table.add_column('Detection ratio', justify='right')

  for domain in embedded_domains:
    analysis_stats =  get_key(domain, 'attributes.last_analysis_stats')
    malicious = analysis_stats['malicious']
    undetected = analysis_stats['undetected']
    table.add_row(
        domain['id'],
        Text(
            '%d / %d' % (malicious, malicious + undetected),
            style='red' if malicious else 'green'))

  if embedded_domains:
    console.print(table)


def print_embedded_urls(embedded_urls):
  embedded_urls = get_key(embedded_urls, 'data', [])

  table = Table(box=box.SIMPLE, show_footer=False, width=CONSOLE_WITH)
  table.add_column('Embedded URLs')
  table.add_column('Detection ratio', justify='right')

  for url in embedded_urls:
    analysis_stats =  get_key(url, 'attributes.last_analysis_stats')
    malicious = analysis_stats['malicious']
    undetected = analysis_stats['undetected']
    table.add_row(
        get_key(url, 'attributes.url'),
        Text(
            '%d / %d' % (malicious, malicious + undetected),
            style='red' if malicious else 'green'))

  if embedded_urls:
    console.print(table)

def print_embedded_ips(embedded_ips):
  embedded_ips = get_key(embedded_ips, 'data', [])

  table = Table(box=box.SIMPLE, show_footer=False, width=CONSOLE_WITH)
  table.add_column('Embedded IP addresses')
  table.add_column('Detection ratio', justify='right')

  for ip in embedded_ips:
    analysis_stats =  get_key(ip, 'attributes.last_analysis_stats')
    malicious = analysis_stats['malicious']
    undetected = analysis_stats['undetected']
    table.add_row(
        ip['id'],
        Text(
            '%d / %d' % (malicious, malicious + undetected),
            style='red' if malicious else 'green'))

  if embedded_ips:
    console.print(table)

#
# scan_file is the function that sends a file to be scanned privately by
# VirusTotal. This is a two step process, in the first step we get an upload
# URL from the /api/v3/private/files/upload_url endpoint, and in the second
# step we upload the file to the received URL. This is required only for files
# larger than 32MB, for smaller files you can post them directly to
# /api/v3/private/files, but we do the two steps process for all files
# regardless of their size for homogeneity.
#
# The response from GET /api/v3/private/files/upload_url looks like:
#
#  {
#    'data': '<the upload URL>'
#  }
#
# The file is POSTed to the obtained URL together with two additional
# arguments: disable_internet and command_line. The first one must be either
# "true" or "false", and indicates whether the file will be detonated in
# a sandbox with Internet connectivity or not, while the second one is the
# string that will be passed to the file as command-line arguments while
# being executed.
#
# This function returns as soon as the file is uploaded and returns the
# identifier for the scan operation. This identifier can be used later on
# for obtaining information about how the scan's progress.
#
def scan_file(client, file_path, sandboxes_disabled=False,
              network_enabled=False, cmd_line='', zip_password=None,
              retention_period_days=None, intercept_tls=False,
              interaction_sandbox='none', interaction_timeout=0):
  """Sends a file to be scanned by VirusTotal privately.

  Args:
    client: An instance of ApiClient.
    file_path: Path to the file that will be.
    sandboxes_disabled: If True, the file is not detonated in sandboxes.
    network_enabled: If True, enables network connectivity while detonating
       the file in a sandbox.
    cmd_line: Command-line arguments passed to the file while executing it.
    zip_password: Password used to unzip the file.
    retention_period_days: Number of days the report and file are kept in VT.
    intercept_tls: If True, enables HTTPS/TLS/SSL interception.
    interaction_sandbox: Sandbox used for live interaction
    interaction_timeout: Number of seconds for live interaction

  Returns:
    The scan identifier.
  """
  # Obtain an upload URL that will be used for uploading the file.
  response = client.get('/api/v3/private/files/upload_url')
  url = response.json()['data']

  file_name = os.path.basename(file_path)

  # POST the file to the previously obtained URL.
  with open(file_path, 'rb') as fin:
    files = {
      'file': (file_name, fin),
    }
    data = {
      'disable_sandbox': str(sandboxes_disabled).lower(),
      'enable_internet':  str(network_enabled).lower(),
      'intercept_tls': str(intercept_tls).lower(),
      'command_line': cmd_line}
    if retention_period_days is not None:
      data['retention_period_days'] = retention_period_days
    if zip_password is not None:
      data['password'] = zip_password
    if interaction_sandbox != 'none':
      data['interaction_sandbox'] = interaction_sandbox
      data['interaction_timeout'] = interaction_timeout

    response = client.post(url, files=files, data=data, timeout=120)
    scan_id = response.json().get('data', {}).get('id')
  return scan_id

#
# This is the function that waits for the completion of private scan. The
# function receives the scan ID (which is obtained from scan_file) and asks
# for the status of the scan every 10 seconds until it gets completed. The
# value returned from this function is the last response obtained from the
# API, which will be a JSON representing a "private_analysis" object, looking
# like this:
#
#  {
#    'data': {
#      'id': <scan ID>,
#      'type': 'private_analysis',
#      'attributes': {
#         'status': 'completed',
#           ... more analysis' attributes
#      }
#    }
#  }
#
# While the scan is in progress, this function shows information about the
# operations that are still in progress, which are returned in the
# "pending_stages" attribute. This attribute is a dictionary where keys are
# strings identifying the components of the scan that are still pending
# and values are strings with more user-friendly description of those
# components. For example:
#
#  {
#    'data': {
#      'id': <scan ID>,
#      'type': 'private_analysis',
#      'attributes': {
#         'status': 'in-progress',
#			    "pending_stages": {
#				      "zenbox": "Sandbox analysis from Zenbox"
#			    }
#      }
#    }
#  }
#
def wait_for_scan(client, scan_id):
  status = None
  start_time = datetime.datetime.now()
  console.print()
  with console.status(
      'Waiting for scan completion, it may take up to 10 minutes...',
       **SPINNER) as spinner:
    while status != 'completed':
      response = client.get('/api/v3/private/analyses/%s' % scan_id)
      attributes = get_key(response.json(), 'data.attributes')
      status = attributes.get('status')
      pending_stage = attributes.get('pending_stages', {})
      if pending_stage:
        msg = 'Waiting for...\n%s\n\n  Time elapsed: %s' % (
          '\n'.join(['  · %s' % s for s in pending_stage.values()]),
          datetime.datetime.now() - start_time)
        # Get live_interaction_url
        sandbox_status = attributes.get('sandbox_status', {})
        live_interaction = [d for d in sandbox_status.values() if 'live_interaction_url' in d]
        for sandbox in live_interaction:
          msg += f'\n  Live interaction url: {sandbox["live_interaction_url"]}'
      else:
        msg = ''
      spinner.update(status=msg)
      time.sleep(10)
  return response

#
# do_intelligence_search_query is a simple function that simply sends a
# VirusTotal Intelligence query and returns the received response.
#
def do_intelligence_search_query(client, query):
  return client.get('/api/v3/intelligence/search', params={'query': query})


def main():
  parser = argparse.ArgumentParser(
      formatter_class=argparse.RawDescriptionHelpFormatter,
      description='''
This script demonstrates the use of the VirusTotal Private Scanning API.

Files passed to this script will be scanned privately, and you'll get back a resume
of the scan results.

You need a VirusTotal API key to use this script, which can be passed either with the
--api-key argument or through the VT_API_KEY environment variable.

The scanned file will be detonated in a sandboxed environment, unless the argument
--sandboxes-disabled is set. This environment doesn't have network connectivity by
default. You can enable network connectivity by using the --network-enabled argument.

Also, you can provide command-line arguments that will be passed to the file while
being executed in a sandboxed environment by using the --cmd-line option.
''')

  parser.add_argument(
      '--api-key',
      help='your VirusTotal API key',
      required=False,
      default='')

  parser.add_argument(
      '--sandboxes-disabled',
      help='don\'t detonate in sandboxes, only apply other characterization tools',
      action='store_true',
      default=False)

  parser.add_argument(
      '--network-enabled',
      help='enables network connectivity while detonating the file in a sandbox',
      action='store_true',
      default=argparse.SUPPRESS)

  parser.add_argument(
      '--intercept-tls',
      help='intercept HTTPS/TLS/SSL to view encypted URLS, hostnames and headers',
      action='store_true',
      default=False)

  parser.add_argument(
      '--interaction-timeout',
      help='number of seconds the live interaction will run for',
      required=False,
      type=int,
      default=60)

  parser.add_argument(
      '--interaction-sandbox',
      help='sandbox used for live interaction',
      required=False,
      choices=['none', 'cape_win', 'cape_linux'],
      default='none')

  parser.add_argument(
      '--cmd-line',
      help='command line passed to the scanned file while executed in sandboxes',
      required=False,
      default='')

  parser.add_argument(
      '--disable-autoupdate',
      help='do not check and autoupdate this script',
      action='store_true',
      default=False)

  parser.add_argument(
      '--zip-password',
      help='password used to unzip the file and scan the contained file',
      required=False,
      default=None)

  parser.add_argument(
      '--destination-dir',
      help='output directory where to store the different artifacts',
      required=False,
      default='.')

  parser.add_argument(
      '--skip-ssl-verification',
      help='skips the SSL certificates verification',
      action='store_true',
      default=False)

  parser.add_argument(
      '--retention-period-days',
      help=(
          'number of days the report and file are kept in VT '
          '(group\'s retention policy preference by default)'),
      required=False,
      type=int,
      default=None)

  # Add file argument as optional
  parser.add_argument(
      'FILE',
      help='path to the file you want to scan')

  args = parser.parse_args()

  if not args.skip_ssl_verification and proxy_check():
    if not Confirm.ask(
        '[red]❯[/red] The SSL certificate received from virustotal.com is not '
        'valid. If you are behind\n  a proxy server this is expected and you '
        'can continue. However, this could also\n  mean that your connection '
        'is being intercepted by an attacker.\n\n  In the future you can use '
        'the --skip-ssl-verification flag for avoiding this question again.'
        '\n\n  Do you want to continue?',
        default=False):
      sys.exit(1)
    console.print()
    args.skip_ssl_verification = True

  if not args.disable_autoupdate and version_update(
      verify_ssl_cert=not args.skip_ssl_verification):
    # Restart the script with the new updated version.
    os.execv(sys.executable, ['python'] + sys.argv)

  if not os.path.isdir(args.destination_dir):
    console.print('The specified path is not a directory')
    sys.exit(1)

  if args.FILE == os.devnull and not args.cmd_line:
    console.print('Must specify command line if no file present')
    parser.print_help()
    sys.exit(1)

  api_key = args.api_key or os.environ.get("VT_APIKEY")

  if not api_key:
    console.print(
      '[bright_red]❯ Your VirusTotal API was not provided either through '
      'the VT_APIKEY environment variable or the --api-key option')

  while True:
    if not api_key:
      api_key = Prompt.ask('\n[red]❯[/red] Enter your API key')
    if re.search('[0-9a-fA-F]{64}', api_key):
      break
    else:
      console.print('  "%s" is not a valid API key' % api_key)
      api_key = None

  client = ApiClient(api_key, verify=not args.skip_ssl_verification)
  mitre_attack_tree = None
  publicly_scanned = False

  if args.zip_password is None:
    # Compute the SHA256 hash of the file.
    try:
      sha256 = compute_sha256(args.FILE)
    except FileNotFoundError as e:
      console.print(e)
      sys.exit(1)

    # Check if the file has being publicly scanned.
    with console.status('Checking if file has been publicly scanned...', **SPINNER):
      try:
        response = client.get('/api/v3/files/%s' % sha256)
        publicly_scanned = response.status_code == 200
      except Exception as e:
        pass

    if publicly_scanned and not Confirm.ask(
        '[red]❯[/red] File is publicly available at... \n\n'
        '[yellow][link={link}]{link}[/link][/yellow]\n\n'
         '...do you want to scan it privately?'
        .format(link='https://www.virustotal.com/gui/file/%s' % sha256), default=False):
      sys.exit(1)
    else:
      console.print()

    # Check if the file has being privately scanned.
    file_info = {}
    with console.status('Checking if file has been privately scanned...', **SPINNER):
      try:
        response = client.get('/api/v3/private/files/%s' % sha256)
        file_info = response.json()
      except Exception as e:
        pass

  # Scan the file if a zip password has been provided, the file wasn't scanned
  # before, or if the user wants to rescan it.
  if args.zip_password is not None or not file_info or Confirm.ask(
      '[red]❯[/red] File already scanned, do you want to scan it again?',
      default=False):

    if args.sandboxes_disabled:
      args.network_enabled = False

    # If the user didn't specified --network-enabled in the command-line, ask
    # if the sandbox should have network connectivity.
    if 'network_enabled' not in args:
      args.network_enabled = Confirm.ask(
          '\n[red]❯[/red] The private scanning logic detonates the file in a '
          'sandboxed environment.\n  Enabling network connectivity in this '
          'environment means that the file may\n  reach out to attacker '
          'infrastructure, which would tip them off. On the\n  other hand, lack '
          'of network connectivity will prevent you from seeing\n  things like '
          'CnC domains, second stage payloads, malware download URLs, etc.\n\n  '
          'Do you want to enable network connectivity?', default=False)

    console.print()

    # Upload the file to VirusTotal and start the scanning.
    with console.status(
        'Uploading file to VirusTotal for private scan...', **SPINNER):
      scan_id = scan_file(
          client,
          args.FILE,
          args.sandboxes_disabled,
          args.network_enabled,
          args.cmd_line,
          args.zip_password,
          args.retention_period_days,
          args.intercept_tls,
          args.interaction_sandbox,
          args.interaction_timeout,
      )

    console.print('[green]✔[/green] File successfully uploaded')
    table = Table(show_header=False, box=None)
    table.add_column(justify='right', min_width=10)
    table.add_column(style='yellow', min_width=10)
    table.add_row('· Scan ID', scan_id)
    console.print(table)

    # Wait for the scan to complete.
    response = wait_for_scan(client, scan_id)
    analysis = response.json()
    sha256 = get_key(analysis, 'meta.file_info.sha256')
    analysis_date = get_key(analysis, 'data.attributes.date')
    file_url = get_key(analysis, 'data.links.item')

    print_unfinished_sandboxes(analysis)

    # Retrieve the final report.
    with console.status('Saving detailed file information...', **SPINNER):
      response = client.get(file_url)
      file_info = response.json()
  else:
    console.print()

  # Change the current directory so we store everything in the specified path.
  os.chdir(args.destination_dir)

  with open('%s.json' % sha256, 'wb') as f:
    f.write(response.content)

  console.print('[green]✔[/green] Detailed file information saved to: %s.json' % sha256)
  console.print()

  with console.status('Getting detailed sandbox report...', **SPINNER):
    try:
      response = client.get('/api/v3/private/files/%s/behaviours' % sha256)
    except Exception:
      pass

    reports = get_key(response.json(), 'data')

  if len(reports) > 0:
    console.print('[green]✔[/green] Found %d recent sandbox reports' % len(reports))
    console.print()

  if len(reports) > 0:
    with console.status('Saving sandbox report...', **SPINNER):
      for report in reports:
        try:
          with open('%s.json' % report['id'], 'w') as f:
            json.dump(report, f, indent=4, separators=(',', ': '))
          console.print(
              '[green]✔[/green] Detailed sandbox report saved to: %s.json' % report['id'],
              highlight=False)
        except Exception as e:
          console.print(
              '[red]✘[/red] Failed to save detailed sandbox report: %s' % report['id'],
              highlight=False)

    console.print()

    behaviour_files = [
        {'status': 'sandbox HTML report',
         'attribute': 'has_html_report',
         'endpoint_suffix': 'html',
         'file_extension': 'html',
         'description': 'Detailed sandbox HTML report'},
        {'status': 'network capture',
         'attribute': 'has_pcap',
         'endpoint_suffix': 'pcap',
         'file_extension': 'pcap',
         'description': 'Network capture (PCAP)'},
        {'status': 'EVTX files',
         'attribute': 'has_evtx',
         'endpoint_suffix': 'evtx',
         'file_extension': 'evtx.zip',
         'description': 'Windows XML Event Log (EVTX)'},
        {'status': 'memdump files',
         'attribute': 'has_memdump',
         'endpoint_suffix': 'memdump',
         'file_extension': 'memdump.zip',
         'description': 'Memdump'}
    ]

    for data in behaviour_files:
      with console.status('Saving {} ...'.format(data['status']), **SPINNER):
        for report in reports:
          if not get_key(report, 'attributes.{}'.format(data['attribute'])):
            continue
          try:
            response = client.get(
                '/api/v3/private/file_behaviours/{}/{}'.format(
                    report['id'], data['endpoint_suffix']))
            with open('{}.{}'.format(report['id'], data['file_extension']), 'wb') as f:
              f.write(response.content)
            console.print('[green]✔[/green] {} saved to: {}.{}'.format(
                data['description'], report['id'],
                data['file_extension']), highlight=False)
          except Exception as e:
            console.print('[red]✘[/red] Failed to save {}: {}'.format(
                data['description'], report['id']), highlight=False)
      console.print()

    with console.status('Saving MITRE ATT&CK information...', **SPINNER):
      try:
        response = client.get(
            '/api/v3/private/files/%s/behaviour_mitre_trees' % sha256).json()
        if response.get('data'):
          with open('%s_mitre_tree.json' % sha256, 'w') as f:
            json.dump(response, f, indent=4, separators=(',', ': '))
            console.print(
                '[green]✔[/green] MITRE ATT&CK tree saved to: %s_mitre_tree.json' % sha256,
                highlight=False)
          mitre_attack_tree = response
      except Exception:
        console.print(
            '[red]✘[/red] Failed to save MITRE ATT&CK information: %s' % sha256,
            highlight=False)

  console.print('\n[green]✔[/green] Scan completed. See results below:\n')

  print_file_identification(file_info)
  print_digital_signature(file_info)
  print_crowdsourced_yara(file_info)
  print_crowdsourced_ids(file_info)
  print_crowdsourced_sigma(file_info)
  print_sandbox_detonation_verdicts(file_info)

  with console.status('Retrieving behaviour information...', **SPINNER):
    try:
      response = client.get('/api/v3/private/files/%s/behaviour_summary' % sha256)
      behaviour_summary = response.json()

      dns_lookups = get_key(behaviour_summary, 'data.dns_lookups', [])
      ip_traffic = get_key(behaviour_summary, 'data.ip_traffic', [])
      http_conversations = get_key(behaviour_summary, 'data.http_conversations', [])
      netloc_info = {}

      for item in dns_lookups:
        hostname = item['hostname']
        response = client.get('/api/v3/domains/%s' % hostname)
        netloc_info[hostname] = {
          'malicious':  get_key(response.json(), 'data.attributes.last_analysis_stats.malicious', 0),
          'harmless':  get_key(response.json(), 'data.attributes.last_analysis_stats.harmless', 0),
          'undetected':  get_key(response.json(), 'data.attributes.last_analysis_stats.undetected', 0),
          'summary': ''}
        domain_creation = get_key(response.json(), 'data.attributes.creation_date')
        if domain_creation:
          domain_creation = datetime.datetime.utcfromtimestamp(domain_creation).strftime('%Y-%m-%d %H:%M:%S')
          netloc_info[hostname]['summary'] = 'Domain creation: %s' % domain_creation

      for item in ip_traffic:
        ip_address = item['destination_ip']
        response = client.get('/api/v3/ip_addresses/%s' % ip_address)
        netloc_info[ip_address] = {
          'malicious':  get_key(response.json(), 'data.attributes.last_analysis_stats.malicious', 0),
          'harmless':  get_key(response.json(), 'data.attributes.last_analysis_stats.harmless', 0),
          'undetected':  get_key(response.json(), 'data.attributes.last_analysis_stats.undetected', 0),
          'summary': ''}
        country = get_key(response.json(), 'data.attributes.country')
        if country:
           netloc_info[ip_address]['summary'] = 'Country: %s' % country

      for item in http_conversations:
        url = item['url']
        response = client.get('/api/v3/urls/%s' % url_id(url))
        netloc_info[url] = {
          'malicious':  get_key(response.json(), 'data.attributes.last_analysis_stats.malicious', 0),
          'harmless':  get_key(response.json(), 'data.attributes.last_analysis_stats.harmless', 0),
          'undetected':  get_key(response.json(), 'data.attributes.last_analysis_stats.undetected', 0),
          'summary': ''}

      print_sandbox_detonation_summary(behaviour_summary, netloc_info)
    except Exception:
      pass

    behaviour_hashes = {}
    try:
      response = client.get(
          '/api/v3/private/files/%s/behaviours' % sha256).json()
      print_sandbox_detonation_details(response)
      for behaviour_report in response.get('data', []):
        behash = get_key(behaviour_report, 'attributes.behash')
        sandbox_name = get_key(behaviour_report, 'attributes.sandbox_name')
        if behash and sandbox_name not in behaviour_hashes:
          behaviour_hashes[sandbox_name] = behash
    except Exception:
      pass

    if mitre_attack_tree:
      print_mitre_attack_tree(mitre_attack_tree)

  with console.status('Looking for similar files...', **SPINNER):
    vhash = get_key(file_info, 'data.attributes.vhash')
    ssdeep = get_key(file_info, 'data.attributes.ssdeep')
    tlsh = get_key(file_info, 'data.attributes.tlsh')
    imphash = get_key(file_info, 'data.attributes.pe_info.imphash')

    rich_pe_header_hash = get_key(file_info,
        'data.attributes.pe_info.rich_pe_header_hash')

    behash_results = {}
    if behaviour_hashes:
      for sandbox_name, behash in behaviour_hashes.items():
        search_results = do_intelligence_search_query(
            client, 'behash:"%s"' % behash)
        behash_results[sandbox_name] = get_key(search_results.json(), 'data', [])

    if vhash:
      search_results = do_intelligence_search_query(client, 'vhash:"%s"' % vhash)
      vhash_list = get_key(search_results.json(), 'data', [])
    else:
      vhash_list = []

    if ssdeep:
      search_results = do_intelligence_search_query(client, 'ssdeep:"%s"' % ssdeep)
      ssdeep_list = get_key(search_results.json(), 'data', [])
    else:
      ssdeep_list = []

    if tlsh:
      search_results = do_intelligence_search_query(client, 'tlsh:"%s"' % tlsh)
      tlsh_list = get_key(search_results.json(), 'data', [])
    else:
      tlsh_list = []

    if imphash:
      search_results = do_intelligence_search_query(client, 'imphash:"%s"' % imphash)
      imphash_list = get_key(search_results.json(), 'data', [])
    else:
      imphash_list = []

    if rich_pe_header_hash:
      search_results = do_intelligence_search_query(client,
          'rich_pe_header_hash:%s' % rich_pe_header_hash)
      rich_pe_header_list = get_key(search_results.json(), 'data', [])
    else:
      rich_pe_header_list = []

    if (vhash_list or tlsh_list or ssdeep or imphash or rich_pe_header_hash
        or behash_results):
      console.rule('[yellow2]Similar files', align='left')

      if vhash_list:
        print_files_table(vhash_list,
            'Files similar by VirusTotal Feature Hash (vhash)',
            'https://www.virustotal.com/gui/search/vhash:"%s"' % vhash)

      if imphash_list:
        print_files_table(imphash_list,
            'Files similar by imphash',
            'https://www.virustotal.com/gui/search/imphash:"%s"' % imphash)

      if behash_results:
        for sandbox_name, results_list in behash_results.items():
          print_files_table(
              results_list,
              'Files similar with %s behaviour report' % sandbox_name,
              'https://www.virustotal.com/gui/search/behash:"%s"' % (
                  behaviour_hashes[sandbox_name]))

      if ssdeep_list:
        print_files_table(ssdeep_list,
            'Files similar by ssdeep context piecewise hashing',
            'https://www.virustotal.com/gui/search/ssdeep:"%s"' % ssdeep)

      if tlsh_list:
        print_files_table(tlsh_list,
            'Files similar by TrendMicro local sensitive hashing (tlsh)',
            'https://www.virustotal.com/gui/search/tlsh:"%s"' % tlsh)

      if rich_pe_header_list:
        print_files_table(rich_pe_header_list,
            'Files similar Rich PE header hash (rich_pe_header_hash)',
            'https://www.virustotal.com/gui/search/rich_pe_header_hash:"%s"' % rich_pe_header_hash)

  with console.status('Looking embedded netlocs...', **SPINNER):
    try:
      response = client.get('/api/v3/private/files/%s/embedded_domains' % sha256)
      embedded_domains = response.json()
      print_embedded_domains(embedded_domains)
    except Exception:
      pass

    try:
      response = client.get('/api/v3/private/files/%s/embedded_urls' % sha256)
      embedded_urls = response.json()
      print_embedded_urls(embedded_urls)
    except Exception:
      pass

    try:
      response = client.get('/api/v3/private/files/%s/embedded_ips' % sha256)
      embedded_ips = response.json()
      print_embedded_ips(embedded_ips)
    except Exception:
      pass


if __name__ == '__main__':
  try:
    main()
  except Exception as e:
   console.print(e)
